import itertools
import logging
import time
import traceback
from typing import Optional, Set, Tuple

from rdfdb.rdflibpatch import contextsForStatement as rp_contextsForStatement
from rdfdb.syncedgraph.readonly_graph import ReadOnlyConjunctiveGraph
from rdfdb.syncedgraph.syncedgraph_base import SyncedGraphBase
from rdflib import ConjunctiveGraph, URIRef
from rdflib.term import Node

log = logging.getLogger("currentstate")

TripleFilter = Tuple[Optional[Node], Optional[Node], Optional[Node]]


class Mgr(object):

    def __init__(self, graph, tripleFilter):
        self._graph = graph
        self._tripleFilter = tripleFilter

    def __enter__(self):
        # this should be a readonly view of the existing
        # graph, maybe with something to guard against
        # writes/patches happening while reads are being
        # done. Typical usage will do some reads on this graph
        # before moving on to writes.

        if True:
            g = ReadOnlyConjunctiveGraph(self._graph)
        else:
            t1 = time.time()
            g = ConjunctiveGraph()
            for s, p, o, c in self._graph.quads(self._tripleFilter):
                g.store.add((s, p, o), c)

            if self._tripleFilter == (None, None, None):
                self.logThisCopy(g, time.time() - t1)

        g.contextsForStatement = lambda stmt: contextsForStatementNoWildcards(g, stmt)
        return g

    def logThisCopy(self, g, sec):
        log.info("copied graph %s statements (%.1f ms) " "because of this:" % (len(g), sec * 1000))
        for frame in traceback.format_stack(limit=4)[:-2]:
            for line in frame.splitlines():
                log.info("  " + line)

    def __exit__(self, type, val, tb):
        return


class CurrentStateGraphApi(SyncedGraphBase):
    """
    mixin for SyncedGraph, separated here because these methods work together
    """

    def currentState(self, context: Optional[URIRef] = None, tripleFilter: TripleFilter = (None, None, None)) -> Mgr:
        """
        a graph you can read without being in an addHandler

        you can save some time by passing a triple filter, and we'll only give you the matching triples
        """
        if context is not None:
            raise NotImplementedError("currentState with context arg")

        return Mgr(self._graph, tripleFilter)

    _reservedSequentials: Optional[Set[URIRef]] = None

    def sequentialUri(self, prefix: URIRef) -> URIRef:
        """
        Prefix URIRef like http://example.com/r- will return
        http://example.com/r-1 if that uri is not a subject in the graph,
        or else http://example.com/r-2, etc
        """
        if self._reservedSequentials is None:
            self._reservedSequentials = set()
        for i in itertools.count(1):
            newUri = URIRef(prefix + str(i))
            if newUri not in self._reservedSequentials and not list(self._graph.triples((newUri, None, None))):
                self._reservedSequentials.add(newUri)
                return newUri
        raise NotImplementedError("never reached")


def contextsForStatementNoWildcards(g, triple):
    if None in triple:
        raise NotImplementedError("no wildcards")
    return rp_contextsForStatement(g, triple)
